# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/19 11:25
from __future__ import print_function

from .logging import it_print
from .stringext import parse_json
from .stringext import to_json
from .timeext import Timer
from .webext import request
from ..functional import deprecated


def print_indent(message):
    """
    缩进打印
    :param message: 待打印消息
    """
    it_print(message, indent=2)


def print_prefix(subject, msg=None):
    """
    以某个消息为前缀打印
    :param subject: 打印内容
    :param msg: 消息前缀
    """
    if msg is not None:
        subject = msg + ' ' + subject
    it_print(subject)


def print_dict(data, msg=None):
    """
    打印字典
    :param data: 数据
    :param msg: 消息提要
    """
    if msg is not None:
        it_print(msg)
    for key in data:
        value = data[key]
        print_indent('{} => {}'.format(key, value))


def json_print(data):
    """
    以json形式打印
    :param data: 待打印数据
    """
    it_print(to_json(data, indent=2, sort_keys=True, ensure_ascii=False))


def print_json(json_string):
    """
    打印json字符串
    :param json_string: json字符串
    :return:
    """
    json_print(parse_json(json_string))


class HttpTest(object):
    """
    Http API测试工具
    """
    debug = True

    def __init__(self, url=None):
        self.url = url

    def set_url(self, host, context, port=80, https=False):
        schema = 'https' if https else 'http'
        self.url = '{}://{}:{}{}'.format(schema, host, port, context)

    def request_url(self, url, data=None, as_json=True, timeout=None):
        # 请求地址
        request_url = url
        if self.debug:
            print_prefix(request_url, "url:")

        # 请求数据
        if data is None:
            data = {}

        if self.debug:
            # 起始时间
            timer = Timer()
            timer.display_start("request at: ")

            # 打印请求数据
            it_print("request:")
            if len(data) != 0:
                json_print(data)

        # 发起请求
        res = request(request_url, data, as_json, timeout)

        if self.debug:
            timer.finish()
            # 打印结果
            it_print("response:")
            if res is not None:
                print_json(res)
            else:
                it_print()

            # 结束时间
            timer.display_finish("finished at: ")
            it_print("time cost: {} s".format(timer.cost()))

            it_print()
        return res

    def request(self, uri, data=None, as_json=True, timeout=None):
        return self.request_url('{}{}'.format(self.url, uri), data, as_json, timeout)


@deprecated(new_fn=json_print)
def print_dict_json(data_dict):
    json_print(data_dict)



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/14 10:53
import sys

import six

from .functional import deprecated

python2 = six.PY2
python3 = six.PY3

if python2:
    C_StandardError = Exception
    C_MAXINT = sys.maxint

    from six.moves import reload_module

    reload_module(sys)
    sys.setdefaultencoding('utf-8')

if python3:
    C_StandardError = BaseException
    C_MAXINT = sys.maxsize


def cse_message(e):
    """
    获取异常消息
    :param e: 异常实例
    :return: 异常消息
    """
    assert isinstance(e, C_StandardError)

    if python2:
        result = e.message
        if result == '':
            result = str(e)
        return result

    if python3:
        if len(e.args) >= 1:
            return e.args[0]
        return ''


def utf8decode(text):
    """
    将text解码为unicode
    :param text: 待解码字符
    :return: 解码后的内容
    """
    return text.decode('utf-8')


def utf8encode(text):
    """
    将text编码为UTF8
    :param text: 待编码内容
    :return: UTF8编码
    """
    return text.encode('utf-8')


TYPE_NAME = 'type' if python2 else 'class'
TYPE_FUNCTION = "<{} 'function'>".format(TYPE_NAME)
TYPE_CLASS = "<{} 'type'>".format(TYPE_NAME)


def compatibility_warning(new_module, *functions):
    """
    用于无警告移动函数
    :param new_module: 新模块
    :param functions: 移动的函数名
    """
    import warnings

    class_count, func_count = 0, 0
    moved_classes, moved_functions = [], []
    for fn in functions:
        type_name = str(type(fn))
        if type_name == TYPE_CLASS:
            class_count += 1
            moved_classes.append(fn.__name__)
        elif type_name == TYPE_FUNCTION:
            func_count += 1
            moved_functions.append(fn.__name__)

    kwargs = {
        'module_name': new_module.__name__,
        'class_count': class_count,
        'func_count': func_count,
        'moved_classes': ','.join(moved_classes),
        'moved_functions': ','.join(moved_functions),

    }

    fmts = ['\n', 'Compatibility warning!\n']
    _fmts = ['  ']
    if class_count > 0:
        _fmts.append('{class_count} classes and ')
    if func_count > 0:
        _fmts.append('{func_count} functions ')
    _fmts.append('have been moved to "{module_name}".\n')
    fmts.append(''.join(_fmts))

    if class_count > 0:
        fmts.append('  classes: "{moved_classes}".\n')
    if func_count > 0:
        fmts.append('  functions: "{moved_functions}".\n')
    fmts.append('  please upgrade to the newest version.')

    warnings.warn(''.join(fmts).format(**kwargs))


@deprecated(new_fn=compatibility_warning)
def happy_move_functions(*args, **kwargs):
    compatibility_warning(*args, **kwargs)



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/9 14:36
from contextlib import contextmanager
from threading import Thread


def deprecated(new_fn=None, version='0.7'):
    """
    标记为弃用 decorator
    :param new_fn: 新的替代函数
    :param version: 被移除版本
    :return 包装函数
    """
    if new_fn is not None:
        assert callable(new_fn)

    def function_wrapper(fn):
        assert callable(fn)

        def wrapper(*args, **kwargs):
            if version is None:
                future = 'the future'
            else:
                future = 'version {}'.format(str(version))

            params = {
                'fn_module': fn.__module__,
                'fn_name': fn.__name__,
                'future': future
            }
            fmts = list(['\n', 'Deprecated warning!\n'])
            fmts.append((
                '  some of your code are using deprecated function: \n'
            ))
            fmts.append((
                '    [{fn_name}] from [{fn_module}].\n'
            ))
            fmts.append((
                '  the deprecated will be removed in {future}.\n'
            ))

            if new_fn is not None:
                params['new_fn_module'] = new_fn.__module__
                params['new_fn_name'] = new_fn.__name__
                fmts.append((
                    '  use [{new_fn_name}] from [{new_fn_module}] instead.'
                    ''))

            import warnings
            message = ''.join(fmts)
            warnings.warn(message.format(**params))

            return fn(*args, **kwargs)

        return wrapper

    return function_wrapper


def async_call(daemon=False, name=None):
    """
    异步调用 decorator
    :param daemon: 是否为守护进程
    :param name: 线程名称
    :return 包装函数
    """

    def function_wrapper(fn):
        """
        函数修饰器
        :param fn: 被调用函数
        """
        assert callable(fn)

        def wrapper(*args, **kwargs):
            t = Thread(target=fn, args=args, kwargs=kwargs)
            t.setDaemon(daemon)
            if name is not None:
                t.setName(name)
            t.start()

        return wrapper

    return function_wrapper


@contextmanager
def none_context():
    yield



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/19 11:03
from .logging import it_print
from ..compat import C_StandardError
from ..functional import deprecated


class InternalError(C_StandardError):

    def __init__(self, *args, **kwargs):
        super(InternalError, self).__init__(*args, **kwargs)


def raise_print(*args, **kwargs):
    """
    print message and raise InternalError
    :param args: see it_print
    :param kwargs: it_print
    """
    _do_print(*args, **kwargs)
    message = kwargs.pop('message', 'Unknown')
    raise InternalError(message)


def exit_print(*args, **kwargs):
    """
    print message and exit
    :param args: see it_print
    :param kwargs: see it_print
    """
    _do_print(*args, **kwargs)
    errno = kwargs.pop('errno', 0)
    exit(int(errno))


def _do_print(*args, **kwargs):
    if 'device' not in kwargs:
        kwargs['device'] = 2

    print_fn = kwargs.pop('print_fn', it_print)
    print_fn(*args, **kwargs)


@deprecated(new_fn=exit_print)
def error_exit(errno=0, error=None):
    """
    print message and exit
    :param errno: error code
    :param error: error message
    """
    if error is not None:
        it_print(error)
    exit(errno)



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/19 11:07
import hashlib
import io
import os
import shutil
import subprocess
import tempfile
from functools import partial

from decorator import contextmanager

from .collext import Iterable
from .stringext import from_json
from .stringext import to_json
from ..compat import utf8encode
from ..functional import deprecated


def abspath(file_obj):
    """
    获取文件绝对路径, 一般在调用文件传__file__对象
    :param file_obj: 文件对象
    :return: 文件的绝对路径
    """
    return os.path.abspath(file_obj)


def absdir(file_obj):
    """
    获取文件所在目录绝对路径, 一般在调用文件传__file__对象
    :param file_obj: 文件对象
    :return: 文件的所在的目录的绝对路径
    """
    _abspath = abspath(file_obj)
    return os.path.dirname(_abspath)


def dirname(path):
    """
    获取路径父路径
    :param path: 路径
    :return: 父路径
    """
    return os.path.dirname(path)


def file_exists(path):
    """
    判断文件是否存在
    :param path: 文件路径
    :return: 文件存在返回True, 否则返回False
    """
    return os.path.islink(path) or os.path.exists(path)


def file_size(path):
    """
    获取文件大小
    :param path: 文件路径
    :return: 文件大小
    """
    return os.path.getsize(path)


def file_lines(path):
    """
    获取文件行数, 通过Shell命令
    :param path: 文件路径
    :return: 文件行数
    """
    if not file_exists(path):
        return -1

    return sum(1 for _ in open(path))


def remove_file(path, ignore_errors=True):
    """
    移除文件(文件及目录)
    :param path: 文件路径
    :param ignore_errors: 是否忽略错误
    :raise 若路径不存在, 则无操作; 若路径既非文件且又非目录则产生ValueError
    """
    if not file_exists(path):
        return

    if os.path.isfile(path) or os.path.islink(path):
        os.remove(path)
    elif os.path.isdir(path):
        shutil.rmtree(path, ignore_errors=ignore_errors)
    else:
        raise ValueError('path should be either a file or directory')


def create_if_not_exists(path):
    """
    如果对应路径不存在的话，则进行创建
    :param path: 文件路径
    :return 若文件已存在且非目录, 抛出ValueError异常
    """
    if os.path.exists(path) and not os.path.isdir(path):
        raise ValueError('path {} exists but is not a directory')

    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)


def concat_path(base_path, sub_path=None, utf8=False):
    """
    拼接路径
    :param base_path: 基本路径
    :param sub_path: 子路径
    :param utf8: 控制是否使用utf8编码(如传输C库)
    :return: 拼接后的路径
    """
    assert base_path is not None
    if sub_path is None:
        path = base_path
    else:
        path = os.path.join(base_path, sub_path)
    return utf8encode(path) if utf8 else path


"""
    拼接路径函数别名
"""
cp = concat_path

"""
    拼接作为C库的路径参数
"""
clp = partial(concat_path, utf8=True)


def is_abspath(file_path):
    """
    判断是否为绝对路径
    :param file_path: 文件路径
    :return: 若为绝对路径返回True, 否则返回False
    """
    cond = file_path is not None and isinstance(file_path, str)
    return cond and file_path.startswith('/')


def basename(file_path):
    return os.path.basename(file_path)


def refine_path(base_path, holder, key):
    """
    修正路径(将holder中key对应的值增加base_path)
    :param base_path: 基本路径
    :param holder: 路径容器
    :param key: 路径键或索引
    """
    if isinstance(holder, dict):
        assert key in holder
    elif isinstance(holder, list):
        assert isinstance(key, int) and 0 <= key < len(holder)
    else:
        s = 'holder should be either dict or list, but got {}'
        raise TypeError(s.format(type(holder)))

    if not is_abspath(holder[key]):
        holder[key] = cp(base_path, holder[key])


def file_md5sum(file_path):
    """
    计算文件的MD5值
    :param file_path: 文件路径
    :return: 文件MD5
    """
    if not os.path.isfile(file_path):
        return
    ctx = hashlib.md5()
    with open(file_path, 'rb') as fp:
        while True:
            b = fp.read(8096)
            if not b:
                break
            ctx.update(b)
    return ctx.hexdigest()


def file_append(path_append_to, path_append_from):
    """
    将path_append_from的文件内容追加文件到path_append_to中
    :param path_append_to: 待追加内容的文件
    :param path_append_from: 待追加内容所在文件
    :return: 追加结果
    """
    kwargs = {
        'path_append_from': path_append_from,
        'path_append_to': path_append_to
    }
    if not file_exists(path_append_to):
        cmd_create_fmt = 'cp {path_append_from} {path_append_to}'
        cmd_create_str = cmd_create_fmt.format(**kwargs)
        try:
            subprocess.check_call(cmd_create_str, shell=True)
        except subprocess.CalledProcessError:
            return False
    else:
        cmd_append_fmt = 'cat {path_append_from} >> {path_append_to}'
        cmd_append_str = cmd_append_fmt.format(**kwargs)
        try:
            subprocess.check_call(cmd_append_str, shell=True)
        except subprocess.CalledProcessError:
            return False
    return True


def load_file_contents(path, pieces=True, strip=True, **openflags):
    """
    读取文件内容
    :param path: 文件路径
    :param pieces: 是否按行返回
    :param strip: 是否对每行进行strip操作
    :return: 若文件不存在返回None, 若可正确读取则返回按行分割的内容列表
    """
    if not file_exists(path):
        return None

    if 'encoding' not in openflags:
        openflags['encoding'] = 'UTF-8'

    with open(path, 'r', **openflags) as fp:
        lines = fp.readlines()
        if strip:
            lines = [x.strip() for x in lines]

        if pieces:
            return lines

        return ''.join([x for x in lines])


def load_json_contents(path):
    contents = load_file_contents(path, pieces=False, strip=True)
    return from_json(contents)


def write_line(wfp, line):
    """
    向文件中写入一行
    :param wfp: 文件写对象
    :param line: 行内容
    """
    if wfp:
        wfp.write(line.strip())
        wfp.write('\n')


def write_pid(path):
    """
    将pid写入到路径所在文件
    :param path: 路径
    """
    with open(path, 'w') as fp:
        fp.write(str(os.getpid()))


def write_file_contents(path, contents, newline=False):
    """
    写入内容至文件
    :param path: 文件路径
    :param contents: 待写入内容
    :param newline: 是否追加新行
    """
    with open(path, 'w') as wfp:
        wfp.write(contents)
        if newline:
            wfp.write('\n')


def write_json_contents(path, data, newline=False):
    """
    写入JSON内容
    :param path: 文件路径
    :param data: 待写入数据
    :param newline: 是否追加新行
    """
    json_str = to_json(data)
    write_file_contents(path, json_str, newline=newline)


def write_iterable_contents(path, iterable_obj, obj2line_func=lambda x: x):
    """
        将可迭代的数据按行的形式写入文件
        :param path: 文件路径
        :param iterable_obj: 可迭代对象
        :param obj2line_func: 将对象映射为行的函数
        """
    with open(path, 'w') as wfp:
        for obj in iterable_obj:
            write_line(wfp, obj2line_func(obj))


@contextmanager
def work_dir(path=None):
    cwd = os.getcwd()
    try:
        if path is not None:
            os.chdir(path)
            yield
    finally:
        os.chdir(cwd)


@contextmanager
def mkdtemp():
    """
    创建临时路径, 并在退出域时删除该路径
    :return: 临时路径
    """
    path = tempfile.mkdtemp()
    create_if_not_exists(path)
    yield path
    shutil.rmtree(path)


@contextmanager
def open_files(*paths, **kwargs):
    """
    批量打开文件
    :param paths: 文件路径
    :param kwargs: 关键字参数(包含打开文件句柄的函数open_fn及相应的函数调用关键字参数)
    :return: 文件句柄
    """
    handles = []
    if not isinstance(paths, Iterable):
        paths = [paths]

    open_fn = kwargs.pop('open_fn', io.open)
    if not callable(open_fn):
        s = 'open_fn should be callable type, but got {}'
        raise TypeError(s.format(type(open_fn)))

    for path in paths:
        handles.append(open_fn(path, **kwargs))

    yield handles

    for handle in handles:
        handle.close()


@deprecated(new_fn=file_append)
def append_to_file(path_append_to, path_append_from):
    return file_append(path_append_to, path_append_from)


@deprecated(new_fn=create_if_not_exists)
def create_if_not_exist_path(file_path):
    create_if_not_exists(file_path)


@deprecated(new_fn=concat_path)
def format_path(base_path, sub_path):
    return concat_path(base_path, sub_path)


@deprecated(new_fn=file_size)
def get_file_size(path):
    return file_size(path)


@deprecated(new_fn=file_lines)
def get_file_lines(path):
    return file_lines(path)


@deprecated(new_fn=write_iterable_contents)
def write_iterable_as_lines(*args, **kwargs):
    write_iterable_contents(*args, **kwargs)



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/19 11:01
from collections import Iterable


def unique_list(items):
    """
    获取无重复的列表
    :param items: 可迭代对象
    :return: 无重复的list对象
    """
    if not isinstance(items, Iterable):
        return items
    return list(set(items))



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018年8月23日 14:26:49
import json

from tornado.httpclient import AsyncHTTPClient
from tornado.web import RequestHandler
from tornado.web import asynchronous

from ..compat import utf8decode
from ..utils.httpclient import json_print
from ..utils.logging import it_print
from ..utils.stringext import to_json
from ..utils.timeext import current_datetime


class WebApplicationHandler(RequestHandler):
    # 错误信息
    error_mapper = dict()

    # 正常响应
    none = 0
    error_mapper[none] = 'none'

    # 无效请求(参数错误)
    invalid_request = 1001
    error_mapper[invalid_request] = 'request param is invalid'

    # 数据未找到
    not_found = 4004
    error_mapper[not_found] = 'not found'

    # 操作不支持
    not_support = 5001
    error_mapper[not_support] = 'not support'

    # 本系统错误
    system_error = 5002
    error_mapper[system_error] = 'system error'

    # 内部服务器错误
    internal_server = 5003
    error_mapper[internal_server] = 'internal server'

    # 调试模式
    debug = False

    # 开发模式
    devel = True

    # 作为后台进程运行
    daemon = True

    @staticmethod
    def setup_config(**kwargs):
        self = WebApplicationHandler
        self.debug = kwargs.pop('debug', self.debug)
        self.devel = kwargs.pop('devel', self.devel)
        self.daemon = kwargs.pop('daemon', self.daemon)

    # 加载为json数据
    def load_request_data(self):
        try:
            body = utf8decode(self.request.body)
            params = json.loads(body) if body != '' else dict()
            # 加入私有属性作为数据，否则params为空，被if not判断为真
            params['_uri'] = self.request.uri
            if self.debug:
                params['request_time'] = current_datetime()
                self.pretty_it_print(params)
        except ValueError:
            return False
        return params

    # 转发请求
    @asynchronous
    def forward(self, url, data=None, callback=None,
                method='POST', timeout=3600):
        if callback is None:
            callback = self.response

        if data is None:
            data = {}

        json_data = to_json(data)
        client = AsyncHTTPClient()
        client.fetch(
            url,
            body=json_data,
            callback=callback,
            method=method,
            request_timeout=timeout
        )

    def success_response(self, data=None):
        self.error_response(self.none, self.error_mapper[self.none], data)

    def error_response(self, error_no=invalid_request,
                       error_desc=None, data=None):
        res = data
        if not data:
            res = dict()
        res['errno'] = error_no
        if error_desc is None:
            error_desc = self.error_mapper[error_no]
        res['error'] = error_desc
        self.__json_response(res)

    def __json_response(self, data):
        if self.debug:
            data['response_time'] = current_datetime()
            self.pretty_it_print(data)
        self.__output_response(to_json(data))

    def __output_response(self, data):
        self.write(data)
        self.finish()

    # 默认响应处理器
    def response(self, response):
        try:
            if self.debug:
                it_print(response.body)
            result = json.loads(response.body)
        except ValueError:
            return self.error_response(error_no=self.system_error)
        return self.success_response(result)

    def data_received(self, chunk):
        pass

    @staticmethod
    def pretty_it_print(data):
        json_print(data)



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/19 11:32
import urllib

import six
from six.moves import xrange

from .collext import Iterable
from .stringext import to_json
from ..compat import C_StandardError


class TimeoutError(Exception):
    """
    # 超时标记接口
    """
    pass


def build_url(host, port, context, uri, schema='http'):
    """
    构建url
    :param host: 主机
    :param port: 端口
    :param context: 上下文
    :param uri: URI路径
    :param schema: 请求模式
    :return: 完整的url
    """
    assert schema == 'http' or schema == 'https'
    return '{}://{}:{}{}{}'.format(schema, host, port, context, uri)


def request(request_url, data=None, as_json=True, timeout=None):
    """
    发送HTTP请求
    :param request_url: 请求url
    :param data: 请求数据
    :param as_json: 是否以json格式发送数据
    :param timeout: 超时时间
    :return: 响应结果或超时异常
    """
    kwargs = dict()
    if timeout is not None:
        kwargs['timeout'] = timeout

    result = None
    if six.PY2:
        from urllib2 import Request
        from urllib2 import build_opener
        from urllib2 import HTTPCookieProcessor
        from urllib2 import URLError

        req = Request(request_url)
        opener = build_opener(HTTPCookieProcessor())
        if data is not None:
            if as_json:
                data = to_json(data, utf8=True)
            else:
                data = urllib.urlencode(data)
        try:
            response = opener.open(req, data, **kwargs)
            result = response.read()
        except URLError as e:
            # e.reason: 输出如，[Errno 61] Connection refused，但Mac和Unix的Errno不一样
            if str(e.reason).find('Connection refused') != -1:
                raise TimeoutError
        except C_StandardError as e:
            if e.message == 'timed out':
                raise TimeoutError
    else:
        from urllib3 import PoolManager
        from urllib3.exceptions import MaxRetryError
        from urllib3.exceptions import NewConnectionError
        from urllib3.exceptions import ReadTimeoutError
        from urllib3.exceptions import RequestError

        http = PoolManager()
        if data is not None:
            if as_json:
                kwargs['body'] = to_json(data, utf8=True)
            else:
                kwargs['fields'] = data
        try:
            response = http.request('POST', request_url, **kwargs)
            result = response.data
        except RequestError as e:
            if isinstance(e, MaxRetryError) \
                    and isinstance(e.reason, NewConnectionError):
                raise TimeoutError
            if isinstance(e, ReadTimeoutError):
                raise TimeoutError
    return result


def fetch_available_port(host='127.0.0.1', used_ports=None, port_range=None):
    """
    获取可用的端口号
    :param host: 主机
    :param used_ports: 端口
    :param port_range: 端口范围
    :return: 端口号
    """
    import socket

    if used_ports is None:
        used_ports = []

    if port_range is None:
        port_range = xrange(50000, 63000)
    else:
        if not isinstance(port_range, Iterable):
            s = 'port_range should be of Iterable type, but got {}'
            raise ValueError(s.format(type(port_range)))

    for port in port_range:
        if port in used_ports:
            continue

        try:
            sk = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sk.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sk.bind((host, port))
            sk.close()
            return port
        except C_StandardError as e:
            if str(e.args[1]) == 'Address already in use':
                continue
    raise C_StandardError('Can\'t get any available port')



from .collext import unique_list
from .exception import InternalError
from .exception import exit_print
from .exception import raise_print
from .fileops import absdir
from .fileops import abspath
from .fileops import basename
from .fileops import concat_path
from .fileops import create_if_not_exists
from .fileops import dirname
from .fileops import file_append
from .fileops import file_exists
from .fileops import file_lines
from .fileops import file_md5sum
from .fileops import file_size
from .fileops import is_abspath
from .fileops import load_file_contents
from .fileops import load_json_contents
from .fileops import mkdtemp
from .fileops import open_files
from .fileops import refine_path
from .fileops import remove_file
from .fileops import work_dir
from .fileops import write_file_contents
from .fileops import write_iterable_contents
from .fileops import write_json_contents
from .fileops import write_line
from .fileops import write_pid
from .httpclient import HttpTest
from .httpclient import json_print
from .httpclient import print_dict
from .httpclient import print_indent
from .httpclient import print_json
from .httpclient import print_prefix
from .logging import disable_print
from .logging import enable_print
from .logging import is_print_enable
from .logging import it_print
from .logging import it_prints
from .logging import set_print_disable
from .logging import set_print_enable
from .monitoring import kill_process
from .stringext import as_json
from .stringext import from_json
from .stringext import md5sum
from .stringext import parse_json
from .stringext import to_json
from .timeext import Timer
from .timeext import current_datetime
from .timeext import current_datetime_str
from .timeext import current_datetime_str
from .timeext import current_datetime_str_s
from .timeext import current_timestamp
from .timeext import timed
from .validation import contain_keys
from .validation import in_range
from .webext import TimeoutError
from .webext import build_url
from .webext import fetch_available_port
from .webext import request



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/11 12:42
from six import exec_


def filter_dict_object(data, *keys, **kwargs):
    """
    获取字典中想要/不要的键, 取决于kwargs参数
    :param data: 字典对象
    :param keys: 需要保留的键
    :param kwargs: 关键字参数
    :return 过滤掉的字典内容
    """
    if not isinstance(data, dict):
        raise TypeError('expected dict type, but got {}'.format(type(data)))
    result = dict()
    keep = False
    if 'keep' in kwargs and kwargs.pop('keep'):
        keep = True
    all_keys = list(data.keys())
    for key in all_keys:
        if keep:
            if key not in keys:
                result[key] = data.pop(key)
        else:
            if key in keys:
                result[key] = data.pop(key)
    return result


def filter_module_methods(module_object, filter_fn=None):
    """
    列出模块方法
    :param module_object: 模块对象
    :param filter_fn: 过滤函数 lambda 方法名: bool
    :return: 满足条件的方法列表
    """

    def _filter_fn_default(x):
        return x.strip() != '' and x[0].islower() and not (x.startswith('_') or x.startswith('__'))

    def _filter_fn_method(x, _):
        exec_('import sys')
        return eval('callable(sys.modules[_.__name__].{})'.format(x))

    if filter_fn is None:
        filter_fn = _filter_fn_default

    def fn(x):
        return filter_fn(x) and _filter_fn_method(x, module_object)

    return list(filter(fn, dir(module_object)))



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/14 16:30
import codecs
import subprocess
import warnings

from .compat import python2
from .utils import current_datetime_str_s
from .utils import file_exists
from .utils import write_file_contents

NOHUP = 'nohup'
BG_MARK = '&'


def _get_log_paths(log_prefix):
    time_suffix = current_datetime_str_s()

    def refine_path(prefix, suffix):
        log_path = '{}.{}'.format(prefix, suffix)
        if file_exists(log_path):
            log_path = '{}.{}'.format(log_path, time_suffix)
        return log_path

    cmd_path = refine_path(log_prefix, 'cmd')
    out_path = refine_path(log_prefix, 'out')
    err_path = refine_path(log_prefix, 'err')

    return cmd_path, out_path, err_path


def shell_invoke(command, **kwargs):
    """
    Shell命令调用
    :param command: 命令
    :param kwargs: 关键字参数

    Args:
        log_prefix: 日志路径前缀
        debug: 是否为调试模式
        daemon: 是否在主线程退出之后仍然运行
        on_error: 用于处理出现错误时的函数
            接收参数 e subprocess.CalledProcessError
    :return: 返回码
    """
    log_prefix = kwargs.pop('log_prefix', None)
    debug = kwargs.pop('debug', True)
    daemon = kwargs.pop('daemon', False)
    on_error = kwargs.pop('on_error', None)
    if on_error is not None and not callable(on_error):
        message = 'on_error must be of callable, but got {}'
        raise TypeError(message.format(type(on_error)))

    command = command.strip()
    if command == '' or command.startswith(NOHUP) or command.endswith(BG_MARK):
        s = (
            'command should be a string not start with nohup '
            'and not end with &, but got "{}"'
        )
        raise ValueError(s.format(command))

    stdout, stderr = None, None
    if log_prefix is not None:
        cmd_path, out_path, err_path = _get_log_paths(log_prefix)
        write_file_contents(cmd_path, command)
        stdout = codecs.open(out_path, mode='w', encoding='utf-8')
        if not debug:
            stderr = codecs.open(err_path, mode='w', encoding='utf-8')
        else:
            stderr = stdout

    if daemon:
        command = '{} {} {}'.format(NOHUP, command, BG_MARK)

    try:
        return subprocess.check_call(
            command, shell=True, stdout=stdout, stderr=stderr
        )
    except subprocess.CalledProcessError as e:
        if on_error:
            on_error(e)
        else:
            raise


def executable_exists(executable):
    """
    检测可执行文件是否存在
    :param executable: 可执行文件
    :return: 检测结果, 若存在，返回True, 否则返回False
    """
    try:
        cmd = 'which {} >/dev/null'.format(executable)
        subprocess.check_call(cmd, shell=True)
    except subprocess.CalledProcessError:
        return False
    return True


def _python_invoke(command, **kwargs):
    """
    Python命令调用
    :param command: 命令
    :param kwargs: 关键字参数

    Args:
        version: Python版本号, 默认为3
        log_prefix: 日志路径前缀
        debug: 是否为调试模式
        daemon: 是否在主线程退出之后仍然运行
        on_error: 错误处理函数
        interpreter: Python解释器
    :return: 返回码
    """
    command = command.strip()
    if command == '' or command.startswith('python'):
        message = 'python command should be a string not start with python'
        raise ValueError(message)

    interpreter = kwargs.pop('interpreter', None)
    if interpreter is None:
        version = kwargs.pop('version', 3)
        if not (version == 2 or version == 3):
            raise ValueError('only support python2 and python3')
        interpreter = 'python{}'.format(version)
        if not executable_exists(interpreter):
            alternative = 'python'
            params = {
                'interpreter': interpreter,
                'alternative': alternative
            }
            message = (
                '{interpreter} not exists, so use {alternative} as '
                'interpreter, ensure that {alternative} is actually '
                'the same as {interpreter}'
            )
            warnings.warn(message.format(**params))
            interpreter = 'python'

    command = '{} -u {}'.format(interpreter, command)
    return shell_invoke(command, **kwargs)


def python2_invoke(command, **kwargs):
    kwargs['version'] = 2
    return _python_invoke(command, **kwargs)


def python3_invoke(command, **kwargs):
    kwargs['version'] = 3
    return _python_invoke(command, **kwargs)


python_invoke = python2_invoke if python2 else python3_invoke



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/8/23 14:29
import ssl

from setuptools import setup, find_packages

ssl._create_default_https_context = ssl._create_unverified_context

setup(
    name='easy_tornado',
    version='0.6.1',
    description='A tornado based web framework package',
    author='Wang Shugen',
    author_email='wsg1107556314@163.com',
    url='https://artifacts.wshugen.cn/python',
    packages=find_packages(),
    install_requires=['tornado', 'decorator', 'six', 'urllib3']
)



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/19 09:30
from __future__ import division
from __future__ import print_function

import functools
import sys

from .stringext import as_json
from ..compat import python3


def _add_indent(lines, space_cnt):
    """
    add blank before message

    :param lines: lines to be operated
    :type lines: Iterable of str

    :param space_cnt: space number
    :type space_cnt: int

    :return: lines with each started with space_cnt blanks
    """
    s = lines.split('\n')
    # don't do anything for single-line stuff
    if len(s) == 1:
        return lines
    first = s.pop(0)
    s = [(space_cnt * ' ') + line for line in s]
    s = '\n'.join(s)
    s = first + '\n' + s
    return s


"""
Whether enable message out or not
"""
_enable = True


def it_print(message='', indent=None, device=1, newline=True, json_fmt=False):
    """
    in time print: print one line to console immediately

    :param message: the message to be printed
    :type message: object

    :param indent: number of blank to be indented, default 0
    :type indent: int

    :param device: stdout -> 1, stderr -> 2, default 1
    :type device: int

    :param newline: whether to append a new line, default True
    :type newline: bool

    :param json_fmt: whether to print as json
    :type json_fmt: bool
    """
    global _enable
    if not _enable or device is None:
        return

    if message is None:
        message = str(None)
    else:
        if not json_fmt:
            if indent is None:
                indent = 0
            message = ' ' * indent + str(message)
        else:
            message = as_json(message, indent=indent)

    if device == 2:
        device = sys.stderr
        message = '\033[0;31m%s\033[0m' % message
    elif device == 1:
        device = sys.stdout

    # a file-like object
    assert hasattr(device, 'write')

    kwargs = dict(file=device,
                  end='\n' if newline else '')
    if python3:
        kwargs['flush'] = True
    print(message, **kwargs)


def it_prints(message='', indent=None, indent_inner=2, device=1, newline=True):
    """
    in time print multiple lines: first indent with indent blanks, then every line is indented with indent_inner blanks
    """
    if message is not None:
        message = _add_indent(message, indent_inner)
    it_print(message, indent=indent, device=device, newline=newline)


def _set_enabled(enable):
    global _enable
    _enable = enable


def is_print_enable():
    """
    get the current printable status
    """
    return _enable


def set_print_enable():
    _set_enabled(True)


def set_print_disable():
    _set_enabled(False)


class disable_print(object):
    def __init__(self):
        self.prev = is_print_enable()

    def __enter__(self):
        set_print_disable()

    def __exit__(self, *args):
        _set_enabled(self.prev)
        return False

    def __call__(self, func):
        @functools.wraps(func)
        def decorate_disable_print(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return decorate_disable_print


class enable_print(object):
    def __init__(self):
        self.prev = is_print_enable()

    def __enter__(self):
        set_print_enable()

    def __exit__(self, *args):
        _set_enabled(self.prev)
        return False

    def __call__(self, func):
        @functools.wraps(func)
        def decorate_disable_print(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return decorate_disable_print



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/19 11:28
from __future__ import division

import time

from .logging import it_print


def current_timestamp():
    """
    获取当前时间戳
    :return: 时间戳
    """
    return time.time()


def current_datetime(timestamp=None):
    """
    获取当前日期
    :param timestamp: 时间戳
    :return: 日期时间格式字符串 形如 2018-11-19 10:20:35
    """
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(timestamp))


def current_datetime_str(timestamp=None):
    """
    获取当前时间戳对应的日期时间字符串
    :param timestamp: 时间戳
    :return: 日期时间字符串 形如 20181119102035
    """
    return time.strftime("%Y%m%d%H%M%S", time.localtime(timestamp))


def current_datetime_str_s(timestamp=None):
    """
    获取当前时间戳对应的以T作为分割的日期、时间字符串
    :param timestamp: 时间戳
    :return: 日期时间字符串 形如 20181119T102035
    """
    return time.strftime("%Y%m%dT%H%M%S", time.localtime(timestamp))


class Timer(object):
    """
    计时器类
    """

    # 无效时间戳标志
    _invalid_ts = -1

    def __init__(self, debug=False):
        self._debug = debug
        self._start_ts = self._invalid_ts
        self._finish_ts = self._invalid_ts
        self.reset()

    def reset(self):
        """
        重新计时
        """
        self._start_ts = time.time()
        self._finish_ts = self._invalid_ts
        if self._debug:
            self.display_start('start at: ')

    @property
    def start_ts(self):
        return self._start_ts

    @property
    def finish_ts(self):
        return self._finish_ts

    @property
    def elapsed(self):
        return time.time() - self.start_ts

    def start(self):
        self.reset()
        return self.start_ts

    def finish(self):
        self._finish_ts = time.time()
        return self.finish_ts

    @property
    def stopped(self):
        return self._finish_ts != self._invalid_ts

    def cost(self):
        self._set_finish()
        return self._finish_ts - self._start_ts

    def state_dict(self):
        return {
            'start_ts': self.start_ts,
            'finish_ts': self.finish_ts
        }

    def load_state_dict_(self, state_dict):
        self._start_ts = state_dict['start_ts']
        self._finish_ts = state_dict['finish_ts']

    def display_start(self, msg=None):
        Timer._display_datetime(self._start_ts, msg)

    def display_finish(self, msg=None):
        self._set_finish()
        Timer._display_datetime(self._finish_ts, msg)

    def display_cost(self, msg=None):
        cost = self.cost()
        prefix = ''
        if msg:
            prefix = '{}'.format(msg)
        it_print('{} cost {}'.format(prefix, self.format(cost)))

    @classmethod
    def format(cls, seconds):
        assert seconds >= 0
        seconds = int(seconds)

        if seconds < 60:
            return '{} seconds'.format(seconds)
        elif 60 <= seconds < 120:
            return cls._format('{} minute {}', seconds, 60)
        elif 120 <= seconds < 3600:
            return cls._format('{} minutes {}', seconds, 60)
        elif 3600 <= seconds < 7200:
            return cls._format('{} hour {}', seconds, 3600)
        elif 7200 <= seconds < 86400:
            return cls._format('{} hours {}', seconds, 3600)
        elif 86400 <= seconds < 172800:
            return cls._format('{} day {}', seconds, 86400)
        else:
            return cls._format('{} days {}', seconds, 86400)

    def _set_finish(self):
        if not self.stopped:
            self.finish()

    @staticmethod
    def _display_datetime(ts, msg=None):
        _tmp_msg = current_datetime(ts)
        if msg is not None:
            if not msg.endswith(' '):
                msg += ' '
            _tmp_msg = msg + _tmp_msg
        it_print(_tmp_msg)

    @classmethod
    def _format(cls, fmt, seconds, factor):
        return fmt.format(seconds // factor, cls.format(seconds % factor))

    def __repr__(self):
        msg = ['start at {}'.format(current_datetime(self.start_ts))]
        if self.stopped:
            msg.append('finished at {}'.format(
                current_datetime(self.finish_ts)
            ))
        return ', '.join(msg)


def timed(description=None, num_pre_blanklines=0, num_suf_blanklines=0):
    """
    计时标记
    :param description: 计时描述
    :param num_pre_blanklines: 输出计时前空行数
    :param num_suf_blanklines: 输出计时后空行数
    :return: 包装函数
    """

    def function_wrapper(fn):
        """
            标记为统计运行时间的函数
            :param fn: 被测函数
            """
        assert callable(fn)

        def wrapper(*args, **kwargs):
            _description = fn.__name__
            if description is not None and not callable(description):
                _description = description

            if num_pre_blanklines > 0:
                [it_print() for _ in range(num_pre_blanklines)]
            timer = Timer()
            timer.display_start('{} start at'.format(_description))
            result = fn(*args, **kwargs)
            timer.display_finish('{} complete at'.format(_description))
            timer.display_cost(_description)
            if num_suf_blanklines > 0:
                [it_print() for _ in range(num_pre_blanklines)]

            return result

        return wrapper

    if callable(description):
        return function_wrapper(fn=description)

    return function_wrapper



# -*- coding: utf-8 -*-
# author: 王树根
# email: wangshugen@ict.ac.cn
# date: 2018/11/19 11:22
import signal
from os import kill

from .fileops import file_exists
from .fileops import load_file_contents
from .fileops import remove_file


def kill_process(pid_path, signum=None):
    """
    向pid_path文件内容中的PID发送信号
    :param pid_path: pid文件路径
    :param signum: 信号值
    """
    if not file_exists(pid_path):
        return None

    if signum is None:
        signum = signal.SIGKILL

    # 读取PID文件
    pid = int(load_file_contents(pid_path)[0].strip())

    # 向该进程发送信号
    try:
        kill(pid, signum)
    except OSError as e:
        if len(e.args) != 2 or e.args[1] != 'No such process':
            raise

    # 移除PID文件
    remove_file(pid_path)
